/*
  sane-troubleshoot: Find out what's wrong with your SANE installation
  Copyright (C) 2003-2005 Henning Meier-Geinitz <henning@meier-geinitz.de>

  This file contains code Copyright (C) 1996, 1997 David Mosberger-Tang 
  and Andreas Beck (sane_strstatus.c).

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <dlfcn.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include "sane.h"
#include "descriptions.h"

#define PROGNAME "sane-troubleshoot"
#define LOGFILE "sane-troubleshoot.log"
#define LOCAL_MAJOR 1
#define LOCAL_MINOR 0
#define LOCAL_BUILD 16
#define DATE "2005-09-08"

#define WEBSITE "http://www.meier-geinitz.de/sane/sts/"
#define UNSUPPORTED "http://www.sane-project.org/contrib.html"
#define MOSTANG "http://www.sane-project.org/"
#define MOSTANG_MAIL "http://www.sane-project.org/mailing-lists.html"
#define ME "Henning Meier-Geinitz <henning@meier-geinitz.de>"

#define NAME_GENERIC "Generic devices"
#define NAME_GENERIC_USB "Generic USB device"
#define NAME_GENERIC_SCSI "Generic SCSI device"
#define NAME_GENERIC_PARPORT "Generic Parport device"
#define NAME_GENERIC_OTHER "Generic other device"


enum label_color_type
  {
    lc_normal = 0,
    lc_warn,
    lc_error
  };

enum check_level_type
  {
    cl_welcome = 0,
    cl_check_sane,
    cl_open_libsane,
    cl_show_devices,
    cl_test_scan,
    cl_test_scan_finish,
    cl_select_from_list,
    cl_check_selection,
    cl_check_usb,
    cl_check_scsi,
    cl_check_parport,
    cl_check_other,
    cl_check_backend,
    cl_load_backend,
    cl_exit
  };


static GtkWidget *main_window;
static GtkWidget *vbox_area;
static GdkColor error_color = {0, 0xa000, 0, 0};
static GdkColor warn_color = {0, 0xa000, 0x5000, 0};
static void *libsane_dl_handle = NULL; /* dlopen handle */
static const SANE_Device ** devices = 0;
static gchar * sane_etc_dir = 0;
static gchar * sane_lib_dir = 0;
static GtkWidget *ok_button, *exit_button;
static enum check_level_type check_level = cl_welcome;
static gint selected_scanner = -1;
static gchar *backend_name;
static GString *manufacturer_name;
static GString *model_name;
static guchar * rgb_buf;
static gint rgb_width;
static gint rgb_height;
static GString *interface_type;


static void check_sane ();
static void open_libsane ();
static void show_devices ();
static void test_scan ();
static void test_scan_finish ();
static void select_from_list ();
static void check_backend ();
static void load_backend ();
static void check_usb ();
static void check_scsi ();
static void check_parport ();
static void check_other ();

SANE_Status (*init_call) (SANE_Int *, SANE_Auth_Callback) = 0;
void (*exit_call) (void);
SANE_Status (*open_call) (SANE_String_Const, SANE_Handle *) = 0;
SANE_Status (*get_devices_call) (const SANE_Device ***, SANE_Bool) = 0;
void (*close_call) (SANE_Handle) = 0;
const SANE_Option_Descriptor * (*get_option_descriptor_call) (SANE_Handle, SANE_Int) = 0;
SANE_Status (*control_option_call) (SANE_Handle, SANE_Int, SANE_Action, void *, SANE_Int *) = 0;
SANE_Status (*get_parameters_call) (SANE_Handle, SANE_Parameters *) = 0;
SANE_Status (*start_call) (SANE_Handle) = 0;
SANE_Status (*read_call) (SANE_Handle, SANE_Byte *, SANE_Int, SANE_Int *) = 0;
void (*cancel_call) (SANE_Handle) = 0;


SANE_String_Const
sane_strstatus (SANE_Status status)
{
  static char buf[80];

  switch (status)
    {
    case SANE_STATUS_GOOD:
      return "Success";

    case SANE_STATUS_UNSUPPORTED:
      return "Operation not supported";

    case SANE_STATUS_CANCELLED:
      return "Operation was cancelled";

    case SANE_STATUS_DEVICE_BUSY:
      return "Device busy";

    case SANE_STATUS_INVAL:
      return "Invalid argument";

    case SANE_STATUS_EOF:
      return "End of file reached";

    case SANE_STATUS_JAMMED:
      return "Document feeder jammed";

    case SANE_STATUS_NO_DOCS:
      return "Document feeder out of documents";

    case SANE_STATUS_COVER_OPEN:
      return "Scanner cover is open";

    case SANE_STATUS_IO_ERROR:
      return "Error during device I/O";

    case SANE_STATUS_NO_MEM:
      return "Out of memory";

    case SANE_STATUS_ACCESS_DENIED:
      return "Access to resource has been denied";

    default:
      /* non-reentrant, but better than nothing */
      sprintf (buf, "Unknown SANE status code %d", status);
      return buf;
    }
}

static void
set_sensible_options (SANE_Handle sane_handle)
{
  SANE_Int num_options;
  const SANE_Option_Descriptor * opt;
  SANE_Status status;
  int i;

  /* get number of options */
  (*control_option_call) (sane_handle, 0, SANE_ACTION_GET_VALUE, &num_options, NULL);

  for (i = 1; i < num_options; i++)
    {
      opt = (*get_option_descriptor_call) (sane_handle, i);
      if (!opt || !opt->name)
	continue;
      /* Try to set a low resolution */
      if (strcmp (opt->name, "resolution") == 0)
	{
	  SANE_Int res = 0;
	  if (opt->constraint_type == SANE_CONSTRAINT_RANGE)
	    res = opt->constraint.range->min;
	  if (opt->type == SANE_TYPE_INT)
	    {
	      if (res < 50)
		res = 50;
	    }
	  else
	    {
	      if (SANE_UNFIX(res) < 50)
		res = SANE_FIX(50);
	    }
	  fprintf (stderr, "--> sane_control_option (\"resolution\", 50)\n");
	  status = (*control_option_call) (sane_handle, i, SANE_ACTION_SET_VALUE, &res, NULL);
	  fprintf (stderr, "<-- sane_control_option (\"resolution\", 50): %s\n", sane_strstatus (status));
	}
      else if (strcmp (opt->name, "br-y") == 0 || strcmp (opt->name, "br-x") == 0)
	{
	  SANE_Int dim = 0;
	  if (opt->unit == SANE_UNIT_MM)
	    dim = 100;
	  else
	    dim = 200;
	  if (opt->type == SANE_TYPE_FIXED)
	    dim = SANE_FIX(dim);
	  fprintf (stderr, "--> sane_control_option (\"br_x/br_y\", 100)\n");
	  status = (*control_option_call) (sane_handle, i, SANE_ACTION_SET_VALUE, &dim, NULL);
	  fprintf (stderr, "<-- sane_control_option (\"br_y/br_y\", 100): %s\n", sane_strstatus (status));
	}
      
    }
}

static void
print_label (gchar *txt, enum label_color_type label_color)
{
  GtkWidget *label = gtk_label_new (txt);
  
  gtk_label_set_selectable (GTK_LABEL (label), TRUE);
  gtk_misc_set_alignment (GTK_MISC(label), 0, 0);
  //gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_RIGHT);
  gtk_label_set_line_wrap (GTK_LABEL (label), TRUE);

  gtk_widget_show (label);

  gtk_box_pack_start (GTK_BOX(vbox_area), label, FALSE, FALSE, 0);
  
  switch (label_color)
    {
    case lc_warn: gtk_widget_modify_fg (label, GTK_STATE_NORMAL, &warn_color); break;
    case lc_error: gtk_widget_modify_fg (label, GTK_STATE_NORMAL, &error_color); break;
    default: break;
    }
}

static void
print_sep ()
{
  GtkWidget *sep = gtk_hseparator_new ();
  gtk_box_pack_start (GTK_BOX(vbox_area), sep, FALSE, FALSE, 0);
  gtk_widget_show (sep);
}

static void
print_not_supported ()
{
  fprintf (stderr, "* Unsupported scanner\n");
  print_label ("As your scanner is not in the list, it is most probably not supported by the "
	       "official version of SANE. It may be supported by an external backend, however. "
	       "I don't know about external backend, so please check the lists at " MOSTANG " "
	       "yourself. If the scanner isn't supported by neither the latest version of SANE "
	       "nor an external backend, a new backend (driver) must be written. You are in a "
	       "good position to write one,"
	       "as you own that device and propably want to get it supported :-). "
	       "Have a look at " UNSUPPORTED " for some hints on how to start. "
	       "Even if you don't want to write a backend, please check if your scanner "
	       "is already listed as \"unsupported\" in the SANE scanners lists (" MOSTANG "). "
	       "If it isn't listed or you have more information, please contact the author "
	       "of this program (" ME ") and tell him the name of your scanner and attach "
	       "the logfile. ", lc_warn);
}


static void
print_sane_devel ()
{
  print_label ("If you can't find out on your own what's going on, please contact the "
	       "sane-devel mailing list for help. Please attach the log file. "
	       "Also describe exactly what you were trying to do and what happened. "
	       "For information on how to subscribe to sane-devel, have a look at "
	       "the SANE mailing list page: " MOSTANG_MAIL, lc_error);
}

static void
print_backend_contact (gchar* backend_name)
{
  GString *ltxt = g_string_new (NULL);
  g_string_printf (ltxt, 
		   "If you can't find out what's wrong yourself, please contact the author of the %s backend "
		   "(see the \"sane-%s\" manual page). If he is not listed in the man page, have a look "
		   "at the AUTHORS file or the source code itsself. If you can't find an address or the backend "
		   "author does not answer, you can ask the sane-devel mailing list for help. "
		   "For information on how to subscribe to sane-devel, have a look at "
		   "the SANE mailing list page: " MOSTANG_MAIL ". "
		   "In either case, please attach the log file. "
		   "Also describe exactly what you were trying to do and what happened. ",
		   backend_name, backend_name);
  print_label (ltxt->str, lc_error);
}

static void
print_wip ()
{
  GString *ltxt = g_string_new (NULL);

  g_string_printf (ltxt,
		   "Not implemented yet. This is work in progress. Please check %s "
		   "for new versions. Happy scanning!", WEBSITE);
  print_label (ltxt->str, lc_error);
}

static void 
bye (GtkWidget *widget, gpointer data)
{
  time_t now = time (NULL);

  widget = widget;
  data = data;

  if (exit_call)
    {
      fprintf (stderr, "--> sane_exit ()\n");
      (*exit_call) ();
      fprintf (stderr, "<-- sane_exit ()\n");
    }
  if (libsane_dl_handle)
    dlclose (libsane_dl_handle);
  if (stderr && check_level != cl_check_sane)
    {
      fprintf (stderr, "Exiting %s", ctime (&now));
      fprintf (stderr, "---------------------------------------------------------------------------\n");
    }
  gtk_main_quit ();
}


static void 
remove_contents (GtkWidget *widget, gpointer data)
{
  data = data;
  gtk_container_remove (GTK_CONTAINER (vbox_area), widget);
}

static void 
ok_callback (GtkWidget *widget, gpointer data)
{

  gtk_container_foreach (GTK_CONTAINER (vbox_area), remove_contents, NULL);

  switch (check_level)
    {
    case cl_check_sane: check_sane (); break; 
    case cl_open_libsane: open_libsane (); break; 
    case cl_show_devices: show_devices (); break; 
    case cl_test_scan: test_scan (); break; 
    case cl_test_scan_finish: test_scan_finish (); break; 
    case cl_select_from_list: select_from_list (); break; 
    case cl_check_selection: 
      if (strcmp (model_name->str, NAME_GENERIC_USB) == 0)
	{
	  print_not_supported ();
	  check_level = cl_check_usb;
	  check_usb ();
	}
      else if (strcmp (model_name->str, NAME_GENERIC_SCSI) == 0)
	{
	  print_not_supported ();
	  check_level = cl_check_scsi;
	  check_scsi ();
	}
      else if (strcmp (model_name->str, NAME_GENERIC_PARPORT) == 0)
	{
	  print_not_supported ();
	  check_level = cl_check_parport;
	  check_parport ();
	}
      else if (strcmp (model_name->str, NAME_GENERIC_OTHER) == 0)
	{
	  print_not_supported ();
	  check_level = cl_check_other;
	  check_other ();
	}
      else
	{
	  check_level = cl_check_backend;
	  check_backend ();
	}
      break;
    case cl_load_backend: load_backend (); break; 
    case cl_check_usb: check_usb (); break; 
    case cl_check_scsi: check_scsi (); break; 
    case cl_check_parport: check_parport (); break; 
    case cl_check_other: check_other (); break; 
    default: bye (widget, data); break; 
    }
}

static void 
scanner_selected (GtkWidget *widget, gpointer data)
{
  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
    {
      selected_scanner = *(int *) data;
      if (selected_scanner >= 0)
	check_level = cl_test_scan;
      else
	check_level = cl_select_from_list;
    }
}

static gboolean
drawing_area_expose (GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
  event = event;
  data = data;

  gdk_draw_rgb_image (widget->window, widget->style->fg_gc[GTK_STATE_NORMAL],
		      0, 0, rgb_width, rgb_height,
		      GDK_RGB_DITHER_MAX, rgb_buf, rgb_width * 3);
  return TRUE;
}


gboolean 
selection_function (GtkTreeSelection *selection, GtkTreeModel *model,
		    GtkTreePath *path, gboolean path_currently_selected,
		    gpointer data)
{
  selection = selection;
  data = data;

  if (!path_currently_selected)
    {
      GtkTreeIter iter;

      gtk_tree_model_get_iter (model, &iter, path);
      if (gtk_tree_model_iter_has_child (model, &iter))
	return FALSE;
      else
	{
	  GtkTreePath *local_path = gtk_tree_path_copy (path);
	  gchar * str_data;

	  gtk_tree_model_get (model, &iter, 0, &str_data, -1);
	  g_string_assign (model_name, str_data);

	  gtk_tree_path_up (local_path);
	  gtk_tree_model_get_iter (model, &iter, local_path);
	  gtk_tree_model_get (model, &iter, 0, &str_data, -1);
	  g_string_assign (manufacturer_name, str_data);

	  gtk_widget_set_sensitive (ok_button, TRUE);
	}
    }
  return TRUE;
}

static void
show_welcome ()
{
  gtk_window_set_title (GTK_WINDOW (main_window), PROGNAME ": Introduction");

  print_label 
    ("Welcome to " PROGNAME ". I'm trying to find out what's wrong with your SANE "
     "(\"Scanner Access Now Easy\") installation. I won't modify any configuration "
     "files myself but will tell you which files to check. "
     "I'll also point you to documentation when appropriate.", lc_normal);
  print_label 
    ("Please connect and turn on your scanner now!", lc_normal);
  print_label 
    ("Everything found out about your SANE installation will be written "
     "to a logfile (\"" LOGFILE "\") in the current directory. "
     "If you run me more than once, the output will be appended to the logfile. "
     "No data will be overwritten. "
     "If you ask the sane-devel mailing list for help, please include the logfile.", lc_normal);
  print_label 
    ("This is version " DATE ". "
     "Currently, this program is in alpha state, so expect all kinds of bugs. If you find any, "
     "please contact the author. Check the website " WEBSITE " for new versions.", lc_normal);
  print_label 
    (PROGNAME " is (C) 2003-2005 by " ME ". This "
     "program is free software; you can redistribute it and/or modify "
     "it under the terms of the GNU General Public License. See COPYING file for "
     "details.", lc_normal);

  check_level = cl_check_sane;
  return;
}


static void
check_sane ()
{
  GDir *dir;
  gchar *dir_name_list[] = {"/usr/local/", "/", "/usr/", "/usr/X11R6/", "/usr/opt/", NULL};
  gchar **dir_names = dir_name_list;
  gchar *dir_name;
  int num_libs = 0;
  GString *found_list = g_string_new (NULL);
  GString *ltxt = g_string_new (NULL);
  time_t now = time (NULL);
  gchar *std_out;

  gtk_window_set_title (GTK_WINDOW (main_window), PROGNAME ": Checking SANE installation");

  fclose (stderr);
  stderr = fopen (LOGFILE, "a");
  if (stderr == 0)
    {
      g_string_printf (ltxt, "The logfile \"%s\" could not be opened for writing. The operating "
	       "system reported the following error: \"%s\". Make sure you have write "
	       "permissions in the current directory and restart %s.\n", LOGFILE, 
	       strerror (errno), PROGNAME);

      print_label (ltxt->str, lc_error);
      check_level = cl_exit;
      gtk_widget_set_sensitive (ok_button, FALSE);
      return;
    }

  fprintf (stderr, "\n+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");
  fprintf (stderr, "%s version %s started %s", PROGNAME, DATE, ctime (&now));
  g_spawn_command_line_sync ("uname -r -s -m", &std_out, NULL, NULL, NULL);
  fprintf (stderr, "System information: %s", std_out);

  print_label 
    ("I'm now checking if your SANE installation is ok. I'll try to find the libraries "
     "and other files associated with SANE. Before going ahead, you should fix any errors "
     "mentioned here.", lc_normal);
  print_sep ();

  
  while ((dir_name = *dir_names++))
    {
      const gchar * file_name;
      GString * libdir_name = g_string_new (NULL);

      g_string_printf (libdir_name, "%slib", dir_name);
      dir = g_dir_open (libdir_name->str, 0, NULL);
      if (!dir)
	continue;
      fprintf (stderr, "Searching for libsane in %s\n", libdir_name->str);
      while ((file_name = g_dir_read_name (dir)))
	{
	  if (strncmp (file_name, "libsane.so", 10) == 0)
	    {
	      if (!sane_lib_dir)
		sane_lib_dir = strdup (libdir_name->str);
	      g_string_append_printf (found_list, "%s\n", libdir_name->str);
	      num_libs++;
	      fprintf (stderr, "Found %s/%s\n", libdir_name->str, file_name);
	      break;
	    }
	}
      g_dir_close (dir);
    }
  switch (num_libs)
    {
    case 0:
      print_label
	("WARNING: I couldn't find the SANE library (libsane.so) in any of the default directories. "
	 "If you have set up SANE with non-standard directories you can ignore this warning, "
	 "but may get into trouble later.", lc_warn);
      fprintf (stderr, "* libsane not found\n");
      break;
    case 1:
      g_string_printf (ltxt, "I have found the SANE library (libsane.so) "
		"at the following location:\n%sThis looks fine for me.", found_list->str);
      print_label (ltxt->str, lc_normal);
      break;
    default:
      g_string_printf (ltxt, "WARNING: Multiple installations of SANE found. "
		"I have found the SANE library (libsane.so) "
		"at %d locations:\n%sThat's bad because the graphical frontends "
		"may use one version and e.g. scanimage may use the other version. "
		"Please make sure that you have installed SANE only once. If you have "
		"installed from source code, remove the installation that came with "
		"your distribution.",
		num_libs, found_list->str);
      print_label (ltxt->str, lc_warn);
      fprintf (stderr, "* More than one SANE installation\n");
      break;
    }

  print_sep ();

  g_string_truncate (found_list, 0);
  dir_names = dir_name_list;
  num_libs = 0;
  while ((dir_name = *dir_names++))
    {
      const gchar * file_name;
      GString * etcdir_name = g_string_new (NULL);

      g_string_printf (etcdir_name, "%setc/sane.d", dir_name);
      dir = g_dir_open (etcdir_name->str, 0, NULL);
      if (!dir)
	continue;
      fprintf (stderr, "Searching for dll.conf in %s\n", etcdir_name->str);
      while ((file_name = g_dir_read_name (dir)))
	{
	  if (strcmp (file_name, "dll.conf") == 0)
	    {
	      if (!sane_etc_dir)
		sane_etc_dir = strdup (etcdir_name->str);
	      g_string_append_printf (found_list, "%s\n", etcdir_name->str);
	      num_libs++;
	      fprintf (stderr, "Found %s/%s\n", etcdir_name->str, file_name);
	      break;
	    }
	}
      g_dir_close (dir);
    }
  switch (num_libs)
    {
    case 0:
      print_label
	("WARNING: I couldn't find the SANE configuration files (dll.conf) in any of the default "
	 "directories. If you have set up SANE with non-standard directories you can ignore "
	 "this warning, but may get into trouble later.", lc_warn);
      fprintf (stderr, "* dll.conf not found\n");
      break;
    case 1:
      g_string_printf (ltxt, "I have found the SANE configuration files "
		"(dll.conf) at the following location:\n%sThis looks fine for me.", found_list->str);
      print_label (ltxt->str, lc_normal);
      break;
    default:
      g_string_printf (ltxt, "WARNING: Multiple SANE configuration files found. "
		"I have found the SANE configuration "
		"files "
		"(dll.conf) "
		"at %d locations:\n%sThese files may be relicts from an old SANE installation. "
		"If you don't get a warning about duplicate libs, this warning can be ignored.",
		num_libs, found_list->str);
      print_label (ltxt->str, lc_warn);
      fprintf (stderr, "* More than one dll.conf\n");
      break;
    }

  print_sep ();

  g_string_truncate (found_list, 0);
  dir_names = dir_name_list;
  num_libs = 0;
  while ((dir_name = *dir_names++))
    {
      const gchar * file_name;
      GString * includedir_name = g_string_new (NULL);

      g_string_printf (includedir_name, "%sinclude/sane", dir_name);
      dir = g_dir_open (includedir_name->str, 0, NULL);
      if (!dir)
	continue;
      fprintf (stderr, "Searching for sane.h in %s\n", includedir_name->str);
      while ((file_name = g_dir_read_name (dir)))
	{
	  if (strcmp (file_name, "sane.h") == 0)
	    {
	      g_string_append_printf (found_list, "%s\n", includedir_name->str);
	      num_libs++;
	      fprintf (stderr, "Found %s/%s\n", includedir_name->str, file_name);
	      break;
	    }
	}
      g_dir_close (dir);
    }
  switch (num_libs)
    {
    case 0:
      print_label 
	("WARNING: I couldn't find the SANE header file (sane/sane.h). You'll only need this file if "
	 "you intend to compile SANE frontends (applications) yourself. In this case, install the "
	 "SANE development package (e.g. libsane-dev). If you just use your distribution's SANE "
	 "installation, you can ignore this warning.", lc_warn);
      fprintf (stderr, "* sane/sane.h not found\n");
      break;
    default:
      g_string_printf (ltxt, "I have found the SANE header file (sane/sane.h) "
		"at the following location:\n%sThis looks fine for me.", found_list->str);
      print_label (ltxt->str, lc_normal);
      break;
    }

  check_level = cl_open_libsane;
  return;
}

static void
open_libsane ()
{
  GString * ltxt = g_string_new (NULL);
  SANE_Status status;
  SANE_Int version_code;
  SANE_Handle sane_handle;
  gboolean found = FALSE, warning = FALSE;
  GString * libsane_name = g_string_new (NULL);

  gtk_window_set_title (GTK_WINDOW (main_window), PROGNAME ": Trying to connect to SANE library");

  print_label
    ("I'm now trying to connect to the SANE library. If I'm successful, "
     "I'll try to open the test backend just to make sure that the "
     "library is installed correctly. If you get any errors, you should exit "
     "and restart me after fixing the problem.", lc_normal);

  print_sep ();

  while (gtk_events_pending ())
    gtk_main_iteration ();

  fprintf (stderr, "trying dlopen\n"); 

  g_string_printf (libsane_name, "%s", "libsane.so");
  if ((libsane_dl_handle = dlopen (libsane_name->str, RTLD_NOW)))
    {
      fprintf (stderr, "dlopen (%s) succeeded\n", libsane_name->str); 
      found = TRUE;
    }
  else
    fprintf (stderr, "dlopen (%s) failed: %s\n", libsane_name->str,  dlerror ()); 

  if (!found)
    {
      g_string_printf (libsane_name, "%s", "libsane.so.1");
      if ((libsane_dl_handle = dlopen (libsane_name->str, RTLD_NOW)))
	{
	  fprintf (stderr, "dlopen (%s) succeeded\n", libsane_name->str); 
	  found = TRUE;
	}
      else
	fprintf (stderr, "dlopen (%s) failed: %s\n", libsane_name->str,  dlerror ()); 
    }

  if (sane_lib_dir)
    {
      if (!found)
	{
	  g_string_printf (libsane_name, "%s/%s", sane_lib_dir, "libsane.so");
	  if ((libsane_dl_handle = dlopen (libsane_name->str, RTLD_NOW)))
	    {
	      fprintf (stderr, "dlopen (%s) succeeded\n", libsane_name->str); 
	      found = TRUE;
	      warning = TRUE;
	    }
	  else
	    fprintf (stderr, "dlopen (%s) failed: %s\n", libsane_name->str,  dlerror ()); 
	}
        if (!found)
	  {
	    g_string_printf (libsane_name, "%s/%s", sane_lib_dir, "libsane.so.1");
	    if ((libsane_dl_handle = dlopen (libsane_name->str, RTLD_NOW)))
	      {
		fprintf (stderr, "dlopen (%s) succeeded\n", libsane_name->str); 
		found = TRUE;
		warning = TRUE;
	      }
	    else
	      fprintf (stderr, "dlopen (%s) failed: %s\n", libsane_name->str,  dlerror ()); 
	  }
    }

  if (!found)
    {
      fprintf (stderr, "*** couldn't dlopen any libsane\n"); 
      g_string_printf (ltxt, 
		       "ERROR: I couldn't open the SANE library libsane.so or libsane.so.1. "
		       "Either you have not installed the SANE backend libraries at all or they "
		       "are at a "
		       "non-standard location. The dynamic linker may not find them. So add "
		       "the directory where the SANE library resides to \"/etc/ld.so.conf\" or set "
		       "the variable LD_LIBRARY_PATH appropriately. Don't forget to run "
		       "\"ldconfig\".");
	      print_label (ltxt->str, lc_error);
	      print_sane_devel ();
	      check_level = cl_exit;
	      gtk_widget_set_sensitive (ok_button, FALSE);
	      return;
    }

  if (warning)
    {
      fprintf (stderr, "* dlopen with full path only\n"); 
      g_string_printf (ltxt, 
		       "WARNING: I was able to open the SANE library (libsane) but had to "
		       "use the full path. If you don't expect to write SANE applications yourself, "
		       "you can most probably ignore this warning. Otherwise add "
		       "\"%s\" to \"/etc/ld.so.conf\" or set the variable LD_LIBRARY_PATH appropriately. "
		       "Don't forget to run \"ldconfig\".", sane_lib_dir);
      print_label (ltxt->str, lc_warn);
      print_sep ();
    }

  init_call = dlsym (libsane_dl_handle, "sane_init");
  if (init_call)
    exit_call = dlsym (libsane_dl_handle, "sane_exit");
  if (exit_call)
    open_call = dlsym (libsane_dl_handle, "sane_open");
  if (open_call)
    close_call = dlsym (libsane_dl_handle, "sane_close");
  if (close_call)
    get_devices_call = dlsym (libsane_dl_handle, "sane_get_devices");
  if (get_devices_call)
    start_call = dlsym (libsane_dl_handle, "sane_start");
  if (start_call)
    get_parameters_call = dlsym (libsane_dl_handle, "sane_get_parameters");
  if (get_parameters_call)
    read_call = dlsym (libsane_dl_handle, "sane_read");
  if (read_call)
    cancel_call = dlsym (libsane_dl_handle, "sane_cancel");
  if (cancel_call)
    get_option_descriptor_call = dlsym (libsane_dl_handle, "sane_get_option_descriptor");
  if (get_option_descriptor_call)
    control_option_call = dlsym (libsane_dl_handle, "sane_control_option");

  if (!init_call || !exit_call || !open_call || !close_call || !get_devices_call
      || !start_call || !get_parameters_call || !read_call || !cancel_call
      || !get_option_descriptor_call || !control_option_call)
    {
      gchar *error_msg = strdup (dlerror ());
      g_string_printf (ltxt, 
		"ERROR: I have opened the SANE library but couldn't get one of the sane_ symbols. "
		"Maybe libsane isn't really the SANE library? Strange. Check your installation. "
		"The error reported by the system was: \"%s\".", error_msg);
      print_label (ltxt->str, lc_error);
      print_sane_devel ();
      fprintf (stderr, "*** dlsym (libsane) failed: %s\n", error_msg); 
      gtk_widget_set_sensitive (ok_button, FALSE);
      check_level = cl_exit;
      return;
    }

  while (gtk_events_pending ())
    gtk_main_iteration ();
  
  fprintf (stderr, "--> sane_init ()\n");
  status = (*init_call) (&version_code, 0);
  fprintf (stderr, "<-- sane_init (): %s\n", sane_strstatus (status));
  
  if (status != SANE_STATUS_GOOD)
    {
      g_string_printf (ltxt,  
		"ERROR: I tried to call sane_init but it returned the following error message: "
		"\"%s\". Check your SANE installation. Maybe the link libsane.so points to a "
		"wrong library?",  sane_strstatus (status));
      print_label (ltxt->str, lc_error);
      print_sane_devel ();
      check_level = cl_exit;
      gtk_widget_set_sensitive (ok_button, FALSE);
      return;
    }

  if (SANE_VERSION_MAJOR (version_code) > LOCAL_MAJOR
      || (SANE_VERSION_MAJOR (version_code) == LOCAL_MAJOR && SANE_VERSION_MINOR (version_code) > LOCAL_MINOR)
      || (SANE_VERSION_MAJOR (version_code) == LOCAL_MAJOR && SANE_VERSION_MINOR (version_code) == LOCAL_MINOR
	  && SANE_VERSION_BUILD (version_code) > LOCAL_BUILD))
    {
      g_string_printf (ltxt, 
		"WARNING: The version of the SANE backends on your system is %d.%d.%d. I don't "
		"know this version yet, so you may want to update %s.", SANE_VERSION_MAJOR (version_code),
		SANE_VERSION_MINOR (version_code), SANE_VERSION_BUILD (version_code), PROGNAME);
      print_label (ltxt->str, lc_warn);
      fprintf (stderr, "* ");
    }
  else if (SANE_VERSION_MAJOR (version_code) < LOCAL_MAJOR
      || (SANE_VERSION_MAJOR (version_code) == LOCAL_MAJOR && SANE_VERSION_MINOR (version_code) < LOCAL_MINOR)
      || (SANE_VERSION_MAJOR (version_code) == LOCAL_MAJOR && SANE_VERSION_MINOR (version_code) == LOCAL_MINOR
	  && (SANE_VERSION_BUILD (version_code) < (LOCAL_BUILD - 3))))
    {
      g_string_printf (ltxt, 
		"WARNING: The version of the SANE backends on your system is %d.%d.%d. "
		"That version is really old. I strongly recommend to update sane-backends "
		"if your scanner is not detected or does not work correctly.",
		SANE_VERSION_MAJOR (version_code),
		SANE_VERSION_MINOR (version_code), SANE_VERSION_BUILD (version_code));
      print_label (ltxt->str, lc_warn);
      fprintf (stderr, "* ");
    }
  else if (SANE_VERSION_MAJOR (version_code) < LOCAL_MAJOR
      || (SANE_VERSION_MAJOR (version_code) == LOCAL_MAJOR && SANE_VERSION_MINOR (version_code) < LOCAL_MINOR)
      || (SANE_VERSION_MAJOR (version_code) == LOCAL_MAJOR && SANE_VERSION_MINOR (version_code) == LOCAL_MINOR
	  && (SANE_VERSION_BUILD (version_code) < LOCAL_BUILD)))
    {
      g_string_printf (ltxt, 
		"WARNING: The version of the SANE backends on your system is %d.%d.%d. "
		"That version is is not the latest available stable version. If "
		"your scanner is not detected or does not work correctly, please "
		"update to at least the latest stable version or even a development "
		"snapshot.",
		SANE_VERSION_MAJOR (version_code),
		SANE_VERSION_MINOR (version_code), SANE_VERSION_BUILD (version_code));
      print_label (ltxt->str, lc_warn);
      fprintf (stderr, "* ");
    }
  else
    {
      g_string_printf (ltxt, 
		"The version of the SANE backends on your system is %d.%d.%d. "
		"If you have trouble with detecting your scanner you may want to try "
		"the latest snapshot of sane-backends. But for now, just go ahead. ",
		SANE_VERSION_MAJOR (version_code),
		SANE_VERSION_MINOR (version_code), SANE_VERSION_BUILD (version_code));
      print_label (ltxt->str, lc_normal);
    }

  fprintf (stderr, "SANE version %d.%d.%d\n", SANE_VERSION_MAJOR (version_code),
	  SANE_VERSION_MINOR (version_code), SANE_VERSION_BUILD (version_code));


  print_sep ();

  while (gtk_events_pending ())
    gtk_main_iteration ();

  fprintf (stderr, "--> sane_open (test:1)\n");
  status = (*open_call) ("test:1", &sane_handle);
  fprintf (stderr, "<-- sane_open (test:1): %s\n", sane_strstatus (status));

  if (status != SANE_STATUS_GOOD)
    {
      g_string_printf (ltxt, 
		"ERROR: I tried to open the test backend but it returned the following error "
		"message: "
		"\"%s\". Check your SANE installation. Maybe sane-backends wasn't installed "
		"completely?",  sane_strstatus (status));
      print_label (ltxt->str, lc_error);
      print_sane_devel ();
      check_level = cl_exit;
      gtk_widget_set_sensitive (ok_button, FALSE);
      return;
    }

  print_label ("I was able to open the test backend. That's fine, you can go ahead now. ", lc_normal);

  fprintf (stderr, "--> sane_close ()\n");
  (*close_call) (sane_handle);
  fprintf (stderr, "<-- sane_close ()\n");

  check_level = cl_show_devices;
}


static void
show_devices ()
{
  GString * ltxt = g_string_new (NULL);
  SANE_Status status;
  int i, *scanner_number;
  GtkWidget *radio_button;

  gtk_window_set_title (GTK_WINDOW (main_window), PROGNAME ": Searching for SANE devices");

  print_label
    ("I'm now displaying the devices found by SANE. "
     "If your scanner is found, "
     "please select it and place a piece of paper (e.g. a photo) at "
     "the top left edge of the scan surface (or the document feeder).", lc_normal); 
  print_label
    ("Your scanner may be listed more than once "
     "at different device files. That's most probably because of links "
     "(e.g. /dev/scanner --> /dev/sg0). That's not a bug. Just select "
     "one of them, it doesn't matter which one.", lc_normal); 
  print_label
    ("If your scanner is not in the list it wasn't found by SANE. "
     "Please select \"My scanner was not found\" in this case. I'll try "
     "to find out the reason when you go ahead.", lc_normal); 

  print_sep ();

  while (gtk_events_pending ())
    gtk_main_iteration ();

  fprintf (stderr, "--> sane_get_devices ()\n");
  status = (*get_devices_call) (&devices, SANE_FALSE);
  fprintf (stderr, "<-- sane_get_devices (): %s\n", sane_strstatus (status));
  if (status != SANE_STATUS_GOOD)
    {
      g_string_printf (ltxt,
		"ERROR: I tried to get the list of SANE devices but SANE returned the following error "
		"message: "
		"\"%s\". That's strange. I don't know what's wrong.",  sane_strstatus (status));
      print_label (ltxt->str, lc_error);
      print_sane_devel ();
      check_level = cl_exit;
      gtk_widget_set_sensitive (ok_button, FALSE);
      return;
    }

  radio_button =
    gtk_radio_button_new_with_label (NULL, "My scanner was not found");
  scanner_number = malloc (sizeof(*scanner_number));
  *scanner_number = -1;
  g_signal_connect (G_OBJECT (radio_button), "toggled",
		    G_CALLBACK (scanner_selected), (gpointer) scanner_number);
  gtk_box_pack_start (GTK_BOX(vbox_area), radio_button, FALSE, FALSE, 0);
  gtk_widget_show (radio_button);

  for (i = 0; devices[i]; ++i)
    {
      GString *scanner_name = g_string_new (NULL);

      if (strcmp (devices[i]->model, "frontend-tester") == 0)
	continue; /* we don't display the test backend to avoid confusion */
      g_string_printf (scanner_name, "%s %s %s at \"%s\"",
		       devices[i]->vendor, devices[i]->model, devices[i]->type,
		       devices[i]->name);
      radio_button =
	gtk_radio_button_new_with_label_from_widget (GTK_RADIO_BUTTON (radio_button), 
						       scanner_name->str);
      scanner_number = malloc (sizeof(*scanner_number));
      *scanner_number = i;
      g_signal_connect (G_OBJECT (radio_button), "toggled",
			G_CALLBACK (scanner_selected), (gpointer) scanner_number);
      gtk_box_pack_start (GTK_BOX(vbox_area), radio_button, FALSE, FALSE, 0);
      gtk_widget_show (radio_button);
      fprintf (stderr, "Found %s %s %s at \"%s\"\n",
	      devices[i]->vendor, devices[i]->model, devices[i]->type,
	      devices[i]->name);
    }

  check_level = cl_select_from_list;
}

static void
test_scan ()
{
  GString *ltxt = g_string_new (NULL);
  GString *dbg_setting = g_string_new (NULL);
  char *position, *scanner_name; 
  GtkWidget * drawing_area, * progress_bar, * hbox;
  SANE_Status status;
  SANE_Handle sane_handle;
  SANE_Int version_code;
  SANE_Parameters params;
  guchar *line_buf;
  gint line_num = 0;

  gtk_window_set_title (GTK_WINDOW (main_window), PROGNAME ": Test scan");


  scanner_name = strdup (devices[selected_scanner]->name);
  backend_name = strdup (scanner_name);
  position = strchr (backend_name, ':');
  if (position)
    *position = 0;

  g_string_printf (dbg_setting, "SANE_DEBUG_%s=255", backend_name);
  dbg_setting = g_string_ascii_up (dbg_setting);

  g_string_printf (ltxt,
		   "So your %s %s was found by SANE. That's good. Your scanner is supported "
		   "by the \"%s\" SANE backend so you should find more information on fine-tuning in "
		   "the \"sane-%s\" manual page. Just run \"man sane-%s\" to view it. "
		   "I'm now trying to scan a small image. This "
		   "will take some time. I'm enabling debug messages (%s), "
		   "so you will be able to see what's going on in the logfile.",
		   devices[selected_scanner]->vendor, devices[selected_scanner]->model,
		   backend_name, backend_name, backend_name, dbg_setting->str);

  print_label (ltxt->str, lc_normal);
  print_sep ();

  fprintf (stderr, "Selected %s %s %s at \"%s\"\n",
	   devices[selected_scanner]->vendor, devices[selected_scanner]->model,
	   devices[selected_scanner]->type, devices[selected_scanner]->name);

  fprintf (stderr, "--> sane_exit ()\n");
  (*exit_call) ();
  fprintf (stderr, "<-- sane_exit ()\n");

  progress_bar = gtk_progress_bar_new ();
  gtk_box_pack_start (GTK_BOX(vbox_area), progress_bar, FALSE, FALSE, 0);
  
  gtk_progress_bar_set_text (GTK_PROGRESS_BAR(progress_bar), "Initializing SANE");
  gtk_widget_show (progress_bar);

  while (gtk_events_pending ())
    gtk_main_iteration ();

  fprintf (stderr, "Setting %s\n", dbg_setting->str);
  putenv (dbg_setting->str);

  fprintf (stderr, "--> sane_init ()\n");
  status = (*init_call) (&version_code, 0);
  fprintf (stderr, "<-- sane_init (): %s\n", sane_strstatus (status));

  gtk_progress_bar_set_text (GTK_PROGRESS_BAR(progress_bar), "Opening scanner");
  while (gtk_events_pending ())
    gtk_main_iteration ();

  fprintf (stderr, "--> sane_open (%s)\n", scanner_name);
  status = (*open_call) (scanner_name, &sane_handle);
  fprintf (stderr, "<-- sane_open (%s): %s\n", scanner_name, sane_strstatus (status));
  if (status != SANE_STATUS_GOOD)
    {
      print_sep ();
      g_string_printf (ltxt, 
		       "ERROR: I tried to call sane_open but it returned the following error message: "
		       "\"%s\". Please check the \"sane-%s\" manual page. Maybe you need to fine-tune "
		       "some settings.", sane_strstatus (status), backend_name);
      print_label (ltxt->str, lc_error);

      g_string_printf (ltxt, 
		       "I'll list some possible causes of that problem: You may need a firmware file but "
		       "it couldn't be found. There may also be communication problems with your device. "
		       "Check syslog (/var/log/messages) for any errors.");
      print_label (ltxt->str, lc_error);

      if (strstr (scanner_name, "libusb"))
	{
	  print_label ("You are using the USB library \"libusb\". You may need to set up "
		       "permissions so you are able to scan as normal user. See the \"libusb\" "
		       "section in the \"sane-usb\" manual page.", lc_error);
	}

      print_backend_contact (backend_name);

      check_level = cl_exit;
      gtk_widget_set_sensitive (ok_button, FALSE);
      return;
    }

  gtk_progress_bar_set_text (GTK_PROGRESS_BAR(progress_bar), "Setting option");
  while (gtk_events_pending ())
    gtk_main_iteration ();

  set_sensible_options (sane_handle);

  gtk_progress_bar_set_text (GTK_PROGRESS_BAR(progress_bar), "Starting scanner");
  while (gtk_events_pending ())
    gtk_main_iteration ();

  fprintf (stderr, "--> sane_start ()\n");
  status = (*start_call) (sane_handle);
  fprintf (stderr, "<-- sane_start (): %s\n", sane_strstatus (status));

  if (status != SANE_STATUS_GOOD)
    {
      print_sep ();
      g_string_printf (ltxt, 
		       "ERROR: I tried to call sane_start but it returned the following error message: "
		       "\"%s\". I think that's a bug in the %s backend. ", sane_strstatus (status), 
		       backend_name);
      print_label (ltxt->str, lc_error);
      print_backend_contact (backend_name);
      check_level = cl_exit;
      gtk_widget_set_sensitive (ok_button, FALSE);
      return;
    }

  gtk_progress_bar_set_text (GTK_PROGRESS_BAR(progress_bar), "Getting parameters");
  while (gtk_events_pending ())
    gtk_main_iteration ();

  fprintf (stderr, "--> sane_get_parameters ()\n");
  status = (*get_parameters_call) (sane_handle, &params);
  fprintf (stderr, "<-- sane_get_parameters (): %s\n", sane_strstatus (status));

  fprintf (stderr, "format: %d, last_frame: %d, bytes_per_line: %d\n", 
	   params.format, params.last_frame, params.bytes_per_line);
  fprintf (stderr, "pixels_per_line: %d, lines: %d, depth: %d\n", 
	   params.pixels_per_line, params.lines, params.depth);
  
  g_string_printf (ltxt, 
		   "Scanning (%d bit %s %d*%d pixels)",
		   params.format == SANE_FRAME_GRAY ? params.depth : params.depth * 3,
		   params.format == SANE_FRAME_GRAY && params.depth == 1 ? "black/white" :
		   params.format == SANE_FRAME_GRAY ? "gray" : "color",
		   params.pixels_per_line, params.lines);

  gtk_progress_bar_set_text (GTK_PROGRESS_BAR(progress_bar), ltxt->str);
  while (gtk_events_pending ())
    gtk_main_iteration ();
  
  if (!params.last_frame || params.lines > 400 || params.pixels_per_line > 600)
    {
      print_label ("I can't scan an image because it's either too big or I don't understand "
		   "the mode (three-pass?). That's most probably not a bug but I'm just not intelligent "
		   "enough. Just go ahead. ", lc_warn);
      fprintf (stderr, "* mode/size error\n");
      check_level = cl_test_scan_finish;
      goto bail_out;
    }

  rgb_width = params.pixels_per_line;
  rgb_height = params.lines;
  rgb_buf = malloc (params.pixels_per_line * params.lines * 3);
  line_buf = malloc (params.bytes_per_line);

  if (!rgb_buf || !line_buf)
    {
      print_label ("Not enough memory for the image buffer. You are really low on memory, aren't you?", lc_warn);
      fprintf (stderr, "* out of memory\n");
      check_level = cl_exit;
      gtk_widget_set_sensitive (ok_button, FALSE);
      goto bail_out;
    }


  hbox = gtk_hbox_new (TRUE, 0);
  gtk_box_pack_start (GTK_BOX(vbox_area), hbox, FALSE, FALSE, 0);
  gtk_widget_show (GTK_WIDGET(hbox));

  drawing_area = gtk_drawing_area_new ();
  gtk_widget_set_size_request (drawing_area, rgb_width, rgb_height);
  gtk_box_pack_start (GTK_BOX(hbox), drawing_area, FALSE, FALSE, 0);
  gtk_signal_connect (GTK_OBJECT (drawing_area), "expose-event",
		      GTK_SIGNAL_FUNC (drawing_area_expose), NULL);


  do 
    {
      SANE_Int bytes_read = 0, total_bytes_read = 0;

      gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR(progress_bar), ((gdouble) line_num + 1) / params.lines);
      while (gtk_events_pending ())
	gtk_main_iteration ();

      do
	{
	  status = (*read_call) (sane_handle, line_buf + total_bytes_read, 
			      params.bytes_per_line - total_bytes_read, &bytes_read);
	  total_bytes_read += bytes_read;
	}
      while (status == SANE_STATUS_GOOD && total_bytes_read < params.bytes_per_line);
      if (params.depth == 16)
	{
	  int i;
	  if (params.format == SANE_FRAME_RGB)
	    {
	      for (i = 0; i < params.pixels_per_line; i++)
		{
		  *(rgb_buf + line_num * params.pixels_per_line * 3 + i * 3) = line_buf[i * 6 + 1];
		  *(rgb_buf + line_num * params.pixels_per_line * 3 + i * 3 + 1) = line_buf[i * 6 + 2 + 1];
		  *(rgb_buf + line_num * params.pixels_per_line * 3 + i * 3 + 2) = line_buf[i * 6 + 4 + 1];
		}
	    }
	  else
	    for (i = 0; i < params.pixels_per_line; i++)
	      {
		*(rgb_buf + line_num * params.pixels_per_line * 3 + i * 3) = line_buf[i * 2 + 1];
		*(rgb_buf + line_num * params.pixels_per_line * 3 + i * 3 + 1) = line_buf[i * 2 + 1];
		*(rgb_buf + line_num * params.pixels_per_line * 3 + i * 3 + 2) = line_buf[i * 2 + 1];
	      }
	}
      else if (params.depth == 8)
	{
	  int i;
	  if (params.format == SANE_FRAME_RGB)
	    memcpy (rgb_buf + line_num * params.bytes_per_line, line_buf, params.bytes_per_line);
	  else
	    for (i = 0; i < params.pixels_per_line; i++)
	      {
		*(rgb_buf + line_num * params.pixels_per_line * 3 + i * 3) = line_buf[i];
		*(rgb_buf + line_num * params.pixels_per_line * 3 + i * 3 + 1) = line_buf[i];
		*(rgb_buf + line_num * params.pixels_per_line * 3 + i * 3 + 2) = line_buf[i];
	      }
	}
      else if (params.depth == 1)
	{
	  int i;
	  int j;
	  for (i = 0; i < params.bytes_per_line; i++)
	    {
	      for (j = 0; j < 8; j++)
		{
		  *(rgb_buf + line_num * params.pixels_per_line * 3 + i * 3 * 8 + j * 3) 
		    = 255 * ((~line_buf[i] >> (7-j)) & 1);
		  *(rgb_buf + line_num * params.pixels_per_line * 3 + i * 3 * 8 + j * 3 + 1)
		    = 255 * ((~line_buf[i] >> (7-j)) & 1);
		  *(rgb_buf + line_num * params.pixels_per_line * 3 + i * 3 * 8 + j * 3 + 2)
		    = 255 * ((~line_buf[i] >> (7-j)) & 1);
		}
	    }
	}
      line_num++;
    }
  while (status == SANE_STATUS_GOOD && line_num < params.lines);

  if (status != SANE_STATUS_GOOD && status != SANE_STATUS_EOF)
    {
      g_string_printf (ltxt, 
		       "I got the following error when reading scan data: %s. I think that's a backend bug, "
		       "but it may also be a bug in %s.", sane_strstatus (status), PROGNAME);
      print_label (ltxt->str, lc_error);
      print_backend_contact (backend_name);
      fprintf (stderr, "sane_read: %s\n", sane_strstatus (status));
      check_level = cl_exit;
      gtk_widget_set_sensitive (ok_button, FALSE);
      goto bail_out;
    }

  gtk_widget_show (drawing_area);
  while (gtk_events_pending ())
    gtk_main_iteration ();

  check_level = cl_test_scan_finish;

 bail_out:

  gtk_progress_bar_set_text (GTK_PROGRESS_BAR(progress_bar), "Stopping scan");
  while (gtk_events_pending ())
    gtk_main_iteration ();
  fprintf (stderr, "--> sane_cancel ()\n");
  (*cancel_call) (sane_handle);
  fprintf (stderr, "<-- sane_cancel ()\n");

  gtk_progress_bar_set_text (GTK_PROGRESS_BAR(progress_bar), "Closing scanner");
  while (gtk_events_pending ())
    gtk_main_iteration ();
  fprintf (stderr, "--> sane_close ()\n");
  (*close_call) (sane_handle);
  fprintf (stderr, "<-- sane_close ()\n");
  gtk_progress_bar_set_text (GTK_PROGRESS_BAR(progress_bar), "Finished scan");
}


static void
test_scan_finish ()
{
  GString *ltxt = g_string_new (NULL);
  
  g_string_printf (ltxt,
		   "Ok, I'm finished now. "
		   "You should be able to scan using a SANE frontend "
		   "like xsane, xscanimage, kooka, quiteinsane or scanimage. For details, "
		   "look at the manual pages and other documentation that comes with these "
		   "frontends.");
  print_label (ltxt->str, lc_normal);
  g_string_printf (ltxt,
		   "If the test image looks bad, e.g. you see stripes or the image is completely black, "
		   "please double-check that this also happens with the above mentioned frontends. "
		   "If it does, it's most probably a bug in the %s backend. ", backend_name);
  print_label (ltxt->str, lc_normal);
  
  print_backend_contact (backend_name);

  print_label ("Please check if your scanner is aready listed as supported at " MOSTANG ". If "
	       "it is listed as unsupported or untested, please contact the backend maintainer "
	       "as mentioned above so the lists can be updated. ", lc_normal);

  print_label ("Thanks for using SANE and " PROGNAME ". Happy scanning!", lc_normal);

  fprintf (stderr, "test_scan_finish: ok\n");
  gtk_widget_set_sensitive (ok_button, FALSE);
  check_level = cl_exit;
}

static void
select_from_list ()
{
  GtkWidget *hbox;
  GtkWidget *scrollbar;
  GList *manufacturer_list = NULL;
  GList *model_list = NULL;
  GtkWidget *tree_view;
  GtkTreeStore *tree_store;
  GtkTreeIter iter, parent;
  GtkCellRenderer *cell;

  gtk_window_set_title (GTK_WINDOW (main_window), PROGNAME ": Select scanner from list");

  fprintf (stderr, "Manual selection from list\n");

  print_label
    ("Ok, so your scanner isn't found by SANE. I'm now generating a list of manufacturers "
     "and models. The devices in this list are known to work with SANE. Please select the "
     "manufacturer "
     "and model name from the lists. If you can't find your device, select "
     "one of the generic devices and the type of interface you use. Generating the list can "
     "take some time.", lc_normal);

  while (gtk_events_pending ())
    gtk_main_iteration ();


  if (!desc_init ())
    {
      print_sep ();
      print_label ("I couldn't read the backend descriptions. Please run "
		   PROGNAME " from the directory where you unpacked it.", lc_error);
      
      gtk_widget_set_sensitive (ok_button, FALSE);
      check_level = cl_exit;
      gtk_widget_set_sensitive (ok_button, FALSE);
      return;
    }

  manufacturer_list = desc_get_manufacturers ();

  tree_store = gtk_tree_store_new (2, G_TYPE_STRING, G_TYPE_STRING);
  tree_view = gtk_tree_view_new ();
  gtk_tree_view_set_model (GTK_TREE_VIEW (tree_view), GTK_TREE_MODEL (tree_store));
  cell = gtk_cell_renderer_text_new ();
  gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (tree_view), 0, "Device", cell,
					       "text", 0, NULL);
  gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (tree_view), 1, "Interface", cell,
					       "text", 1, NULL);

  gtk_tree_store_append (tree_store, &iter, NULL);
  parent = iter;
  gtk_tree_store_set (tree_store, &iter, 0, NAME_GENERIC, -1);
  gtk_tree_store_append (tree_store, &iter, &parent);
  gtk_tree_store_set (tree_store, &iter, 0, NAME_GENERIC_USB, 1, "USB", -1);
  gtk_tree_store_append (tree_store, &iter, &parent);
  gtk_tree_store_set (tree_store, &iter, 0, NAME_GENERIC_SCSI, 1, "SCSI", -1);
  gtk_tree_store_append (tree_store, &iter, &parent);
  gtk_tree_store_set (tree_store, &iter, 0, NAME_GENERIC_PARPORT, 1, "Parport", -1);
  gtk_tree_store_append (tree_store, &iter, &parent);
  gtk_tree_store_set (tree_store, &iter, 0, NAME_GENERIC_OTHER, 1, "Other", -1);

  while (manufacturer_list)
    {
      model_list = desc_get_models (manufacturer_list->data);
      if (model_list)
	{
	  gtk_tree_store_append (tree_store, &iter, NULL);
	  parent = iter;
	  gtk_tree_store_set (tree_store, &iter, 0, manufacturer_list->data, -1);
	  while (model_list)
	    {
	      gtk_tree_store_append (tree_store, &iter, &parent);
	      gtk_tree_store_set (tree_store, &iter, 
				  0, ((model_record_entry *) model_list->data)->name,
				  1, ((model_record_entry *) model_list->data)->interface,
				  -1);
	      model_list = g_list_next (model_list);
	    }
	}
      manufacturer_list = g_list_next (manufacturer_list);
    }


  hbox = gtk_hbox_new (SANE_FALSE, 0);
  scrollbar  = gtk_vscrollbar_new (gtk_tree_view_get_vadjustment (GTK_TREE_VIEW (tree_view)));
  gtk_box_pack_start (GTK_BOX(vbox_area), hbox, TRUE, TRUE, 0);
  gtk_widget_show (hbox);
  gtk_widget_show (scrollbar);
  gtk_box_pack_start (GTK_BOX(hbox), scrollbar, FALSE, FALSE, 0);
  gtk_box_pack_start (GTK_BOX(hbox), tree_view, FALSE, FALSE, 0);
  gtk_widget_set_size_request (tree_view, -1, 300);
  gtk_widget_show (tree_view);
  
  gtk_tree_selection_set_select_function (gtk_tree_view_get_selection (GTK_TREE_VIEW(tree_view)),
					  selection_function, (gpointer) NULL, NULL);
  gtk_widget_set_sensitive (ok_button, FALSE);
  check_level = cl_check_selection;
}

static void
check_backend ()
{
  GString *ltxt = g_string_new (NULL);
  model_record_entry *model_record;
  GString * dll_conf_path = g_string_new (sane_etc_dir);
  gchar *dll_file;
  gchar ** lines;
  gboolean found = FALSE;
  int i;

  interface_type = g_string_new (NULL);

  gtk_window_set_title (GTK_WINDOW (main_window), PROGNAME ": Check backend");

  model_record = desc_get_model (manufacturer_name->str, model_name->str);

  backend_name = g_ascii_strdown (model_record->be->name, -1);
  
  fprintf (stderr, "Selected %s %s -> %s backend\n", manufacturer_name->str, model_name->str,
	   backend_name);
  
  g_string_printf (ltxt, "Your %s %s should be supported by the SANE \"%s\" backend. At least that's what "
		   "our lists say. As it wasn't found, I guess something is wrong with your installation "
		   "or kernel setup. I'll try to find out more details later. ", manufacturer_name->str,
		   model_name->str, backend_name);
  print_label (ltxt->str, lc_normal);

  g_string_printf (ltxt, "You will find more information about the %s backend in the %s manual page. ",
		   backend_name, model_record->be->manpage);
  fprintf (stderr, "manpage: %s\n", model_record->be->manpage);

  if (model_record->be->url)
    {
      g_string_append_printf (ltxt, "The website (or email address) for the %s backend is "
			      "%s. ", backend_name, model_record->be->url->name);
      fprintf (stderr, "URL: %s\n",  model_record->be->url->name);
    }
  if (model_record->be->comment)
    {
      g_string_append_printf (ltxt, "The backend is attached with the following comment: \"%s\".",
			      model_record->be->comment);
      fprintf (stderr, "comment: %s\n", model_record->be->comment);
    }
  print_label (ltxt->str, lc_normal);
  
  if (model_record->status == status_untested)
    {
      print_label ("Your scanner is labelled as \"untested\". That means that the backend "
		   "maintainer isn't sure if the scanner works.", lc_warn);
      print_backend_contact (backend_name);
      fprintf (stderr, "* Backend is untested\n");
    }

  g_string_append_printf (dll_conf_path, "/dll.conf");
  
  if (!g_file_get_contents (dll_conf_path->str, &dll_file, NULL, NULL))
    {
      fprintf (stderr, "Couldn't open %s\n", dll_conf_path->str);
      g_string_printf (ltxt, "Huh? I wasn't able to open dll.conf (%s). Most probably a bug"
		       "in %s.", dll_conf_path->str, PROGNAME);
      print_label (ltxt->str, lc_error);
      check_level = cl_exit;
      gtk_widget_set_sensitive (ok_button, FALSE);
      return;
    }

  print_sep ();

  lines = g_strsplit (dll_file, "\n", 0);

  for (i = 0; lines[i]; i++)
    {
      gchar *start = lines[i];
      gchar *end;

      while (g_ascii_isspace (*start))
	start++;
      if (*start == '#')
	continue;
      end = start;
      while (*end && !g_ascii_isspace (*end))
	end++;
      *end = '\0';
      if (strcmp (start, backend_name) == 0)
	found = TRUE;
    }

  if (!found)
    {
      fprintf (stderr, "*** Backend %s not found in %s\n", backend_name,
	       dll_conf_path->str);	       
      g_string_printf (ltxt, "The backend \"%s\" is not listed in %s. Maybe it is commented out? "
		       "Please edit this file (as root) and add %s to it. Or, if it's already "
		       "there, remove the comment sign (#). After that, restart me.", backend_name,
		       dll_conf_path->str, backend_name);
      print_label (ltxt->str, lc_error);
      check_level = cl_exit;
      gtk_widget_set_sensitive (ok_button, FALSE);
      return;
    }
  else
    {
      fprintf (stderr, "Backend %s found in %s\n", backend_name,
	       dll_conf_path->str);	       
      g_string_printf (ltxt, "I've found the backend \"%s\" in %s. That's ok, please go ahead.", backend_name,
		       dll_conf_path->str);
      print_label (ltxt->str, lc_normal);
    }

  g_string_assign (interface_type, model_record->interface);

  if (exit_call)
    {
      fprintf (stderr, "--> sane_exit ()\n");
      (*exit_call) ();
      fprintf (stderr, "<-- sane_exit ()\n");
    }
  if (libsane_dl_handle)
    dlclose (libsane_dl_handle);
  exit_call = NULL;
  libsane_dl_handle = NULL;

  check_level = cl_load_backend;
  //  gtk_widget_set_sensitive (ok_button, FALSE);
}

static void
load_backend ()
{
  GString *ltxt = g_string_new (NULL);
  SANE_Int version_code;
  GString *dbg_setting = g_string_new (NULL);
  SANE_Status status;
  GString * libsane_name = g_string_new (NULL);
  gboolean found = FALSE;
  int i;

  gtk_window_set_title (GTK_WINDOW (main_window), PROGNAME ": Load backend");

  g_string_printf (ltxt, "Now I'm trying to load the %s backend and enable debugging. "
		   "So you can check the logfile for any errors. ", backend_name);
  print_label (ltxt->str, lc_normal);

  print_backend_contact (backend_name);

  print_sep ();

  while (gtk_events_pending ())
    gtk_main_iteration ();

  if (sane_lib_dir)
    {
      g_string_printf (libsane_name, "%s/sane/libsane-%s.so", sane_lib_dir, backend_name);
      if ((libsane_dl_handle = dlopen (libsane_name->str, RTLD_NOW)))
	    {
	      fprintf (stderr, "dlopen (%s) succeeded\n", libsane_name->str); 
	      found = TRUE;
	    }
	  else
	    fprintf (stderr, "dlopen (%s) failed: %s\n", libsane_name->str,  dlerror ()); 
        if (!found)
	  {
	    g_string_printf (libsane_name, "%s/sane/libsane-%s.so.1", sane_lib_dir, backend_name);
	    if ((libsane_dl_handle = dlopen (libsane_name->str, RTLD_NOW)))
	      {
		fprintf (stderr, "dlopen (%s) succeeded\n", libsane_name->str); 
		found = TRUE;
	      }
	    else
	      fprintf (stderr, "dlopen (%s) failed: %s\n", libsane_name->str,  dlerror ()); 
	  }
    }

  if (!found)
    {
      fprintf (stderr, "*** couldn't dlopen sane/libsane-%s.so\n", libsane_name->str); 
      g_string_printf (ltxt, 
		       "ERROR: I couldn't open the %s backend (%s). Maybe that backend hasn't been "
		       "installed? Double-check that your SANE version is current.", backend_name,
		       libsane_name->str);
	      print_label (ltxt->str, lc_error);
	      check_level = cl_exit;
	      gtk_widget_set_sensitive (ok_button, FALSE);
	      return;
    }


  init_call = dlsym (libsane_dl_handle, "sane_init");
  if (init_call)
    exit_call = dlsym (libsane_dl_handle, "sane_exit");
  if (exit_call)
    open_call = dlsym (libsane_dl_handle, "sane_open");
  if (open_call)
    close_call = dlsym (libsane_dl_handle, "sane_close");
  if (close_call)
    get_devices_call = dlsym (libsane_dl_handle, "sane_get_devices");
  if (get_devices_call)
    start_call = dlsym (libsane_dl_handle, "sane_start");
  if (start_call)
    get_parameters_call = dlsym (libsane_dl_handle, "sane_get_parameters");
  if (get_parameters_call)
    read_call = dlsym (libsane_dl_handle, "sane_read");
  if (read_call)
    cancel_call = dlsym (libsane_dl_handle, "sane_cancel");
  if (cancel_call)
    get_option_descriptor_call = dlsym (libsane_dl_handle, "sane_get_option_descriptor");
  if (get_option_descriptor_call)
    control_option_call = dlsym (libsane_dl_handle, "sane_control_option");

  if (!init_call || !exit_call || !open_call || !close_call || !get_devices_call
      || !start_call || !get_parameters_call || !read_call || !cancel_call
      || !get_option_descriptor_call || !control_option_call)
    {
      gchar *error_msg = strdup (dlerror ());
      g_string_printf (ltxt, 
		"ERROR: I have opened the SANE library but couldn't get one of the sane_ symbols. "
		"Maybe %s isn't really the SANE library? Strange. Check your installation. "
		"The error reported by the system was: \"%s\".", libsane_name->str, error_msg);
      print_label (ltxt->str, lc_error);
      print_sane_devel ();
      fprintf (stderr, "*** dlsym (%s) failed: %s\n", libsane_name->str, error_msg); 
      gtk_widget_set_sensitive (ok_button, FALSE);
      check_level = cl_exit;
      return;
    }

  g_string_printf (dbg_setting, "SANE_DEBUG_%s=255", backend_name);
  dbg_setting = g_string_ascii_up (dbg_setting);
  
  fprintf (stderr, "Setting %s\n", dbg_setting->str);
  putenv (dbg_setting->str);

  fprintf (stderr, "--> sane_init ()\n");
  status = (*init_call) (&version_code, 0);
  fprintf (stderr, "<-- sane_init (): %s\n", sane_strstatus (status));

  if (status == SANE_STATUS_GOOD)
    {
      fprintf (stderr, "--> sane_get_devices ()\n");
      status = (*get_devices_call) (&devices, SANE_FALSE);
      fprintf (stderr, "<-- sane_get_devices (): %s\n", sane_strstatus (status));
      if (status == SANE_STATUS_GOOD)
	{
	  if (devices[0])
	    {
	      g_string_printf (ltxt,  
			       "WARNING: Strange, when I called sane_get_devices (), the following "
			       "devices are listed:\n");
	      for (i = 0; devices[i]; i++)
		{
		  g_string_append_printf (ltxt, "%s %s %s at \"%s\"\n",
					  devices[i]->vendor, devices[i]->model, devices[i]->type,
					  devices[i]->name);
		  fprintf (stderr, "* found: %s %s %s at \"%s\"\n",
			   devices[i]->vendor, devices[i]->model, devices[i]->type,
			   devices[i]->name);
		}
	      g_string_append_printf (ltxt,  
				      "Are you sure that they didn't show up in the list "
				      "where you decided that \"My scanner was not found\"? "
				      "If yes, there is problem with loading the dynamic libraries. "
				      "If you have more than one scanner that's supported by the "
				      "%s backend, please only connect one of them at the same time for "
				      "testing.", backend_name);

	      print_label (ltxt->str, lc_error);
	      check_level = cl_exit;
	      gtk_widget_set_sensitive (ok_button, FALSE);
	      return;
	    }
	  
	  if (strstr (interface_type->str, "USB"))
	    {
	      check_level = cl_check_usb;
	    }
	  else if (strstr (interface_type->str, "SCSI"))
	    {
	      check_level = cl_check_scsi;
	    }
	  else if (strstr (interface_type->str, "Parport"))
	    {
	      check_level = cl_check_parport;
	    }
	  else
	    {
	      check_level = cl_check_other;
	    }


	  g_string_printf (ltxt, "I was able to load the backend and call some of its functions, so "
			   "your SANE installation seems to be ok. I think your scanner is connected "
			   "through a %s interface, so I'll see if I can find out more about that "
			   "interface when you go ahead.", interface_type->str);
	  print_label (ltxt->str, lc_normal);

	  fprintf (stderr, "test succeeded, interface_type = %s\n", interface_type->str);
	  return;
	}
    }
    
  
  g_string_printf (ltxt,  
		"ERROR: I tried to call one of the SANE functions but it returned the following error message: "
		"\"%s\". I'm not sure what that means. Check your SANE installation.",  sane_strstatus (status));
  print_label (ltxt->str, lc_error);
  print_sane_devel ();
  check_level = cl_exit;
  gtk_widget_set_sensitive (ok_button, FALSE);
  return;
}


static void
check_usb ()
{
  gtk_window_set_title (GTK_WINDOW (main_window), PROGNAME ": Search USB devices");

  print_wip ();
  check_level = cl_exit;
  gtk_widget_set_sensitive (ok_button, FALSE);
}

static void
check_scsi ()
{
  gtk_window_set_title (GTK_WINDOW (main_window), PROGNAME ": Search SCSI devices");
  print_wip ();
  check_level = cl_exit;
  gtk_widget_set_sensitive (ok_button, FALSE);
}

static void
check_parport ()
{
  gtk_window_set_title (GTK_WINDOW (main_window), PROGNAME ": Search Parallel port devices");

  print_wip ();
  check_level = cl_exit;
  gtk_widget_set_sensitive (ok_button, FALSE);
}

static void
check_other ()
{
  gtk_window_set_title (GTK_WINDOW (main_window), PROGNAME ": Search other devices");

  print_wip ();
  check_level = cl_exit;
  gtk_widget_set_sensitive (ok_button, FALSE);
}


int 
main (int argc, char *argv[])
{
  GtkWidget *vbox, *button_box;
  GtkWidget *sep;
  GtkWidget *frame;

  gtk_init (&argc, &argv);

  manufacturer_name = g_string_new (NULL);
  model_name = g_string_new (NULL);
  
  main_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  //  gtk_window_set_default_size (GTK_WINDOW(main_window), 800, 600);
  g_signal_connect (G_OBJECT (main_window), "delete_event",
		    G_CALLBACK (bye), NULL);
  gtk_container_set_border_width (GTK_CONTAINER (main_window), 10);

  gtk_window_set_position (GTK_WINDOW (main_window), GTK_WIN_POS_CENTER);

  ok_button = gtk_button_new_from_stock (GTK_STOCK_OK);
  exit_button = gtk_button_new_from_stock (GTK_STOCK_QUIT);
  vbox_area = gtk_vbox_new (FALSE, 5);
  vbox = gtk_vbox_new (FALSE, 0);
  button_box = gtk_hbutton_box_new ();
  sep = gtk_hseparator_new ();
  frame = gtk_frame_new (NULL);

  g_signal_connect (G_OBJECT (exit_button), "clicked",
		    G_CALLBACK (bye), NULL);
  g_signal_connect (G_OBJECT (ok_button), "clicked",
		    G_CALLBACK (ok_callback), vbox);

  gtk_container_add (GTK_CONTAINER (main_window), frame);
  gtk_container_add (GTK_CONTAINER (frame), vbox);
  gtk_box_pack_start (GTK_BOX(vbox), vbox_area, FALSE, FALSE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (vbox_area), 10);

  gtk_box_pack_end (GTK_BOX(vbox), button_box, FALSE, FALSE, 0);
  gtk_box_pack_end (GTK_BOX(vbox), sep, FALSE, FALSE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (button_box), 10);
  gtk_container_add (GTK_CONTAINER (button_box), ok_button);
  gtk_container_add (GTK_CONTAINER (button_box), exit_button);

  gtk_widget_show (frame);
  gtk_widget_show (vbox);
  gtk_widget_show (vbox_area);
  gtk_widget_show (button_box);
  gtk_widget_show (sep);
  gtk_widget_show (exit_button);
  gtk_widget_show (ok_button);
  gtk_widget_show (main_window);
  gtk_window_set_position (GTK_WINDOW (main_window), GTK_WIN_POS_CENTER_ALWAYS);

  show_welcome ();
  
  gtk_main ();
  return 0;
}
